""" Decentralized Riemannian Stochastic Gradient Descent over Stiefel manifold. """


import math
import numpy as np
from mpi4py import MPI

from Log_File.Log import Log


comm = MPI.COMM_WORLD
size = MPI.COMM_WORLD.Get_size()
rank = MPI.COMM_WORLD.Get_rank()

class DRSGD:

    def __init__(self, 
                 local_obj,
                 manifold,
                 consensus,
                 initial_guess, 
                 batch_size,
                 step_size, 
                 consensus_step,
                 termination_condition = None,
                 log = None, 
                 if_compute_obj_value = False):
    
        self.local_obj = local_obj
        self.manifold = manifold
        self.consensus = consensus
        self.initial_guess = initial_guess
        self.batch_size = batch_size
        self.step_size = step_size
        self.consensus_step = consensus_step
        self.termination_condition = termination_condition
        self.log = log
        self.if_compute_obj_value = if_compute_obj_value

    def minimize(self):

        iter = 0

        mi = self.local_obj.sample_size
        idx = np.arange(mi)

        assert isinstance(self.batch_size, int) and self.batch_size > 0, 'batch size should be a positive integer'

        max_iter = self.termination_condition.get('iteration')
        if max_iter is None:
            max_iter = 10**4
        tol_grad = self.termination_condition.get('gradient_tolerance')
        if tol_grad is None:
            tol_grad = 1e-4
        tol_cons = self.termination_condition.get('consensus_tolerance')
        if tol_cons is None:
            tol_cons = 1e-10
        stop_criteria = iter < max_iter

        step_size_cons = self.step_size.get('consensus')
        if step_size_cons is None:
            step_size_cons = 1
        step_size_grad = self.step_size.get('gradient')
        if step_size_grad is None:
            step_size_grad = 0.0001
        step_size_type = self.step_size.get('type')
        if step_size_type is None:
            step_size_type = 'BB'

        x = self.initial_guess
        fval, grad = self.local_obj.objective(x, idx, self.if_compute_obj_value)
        rgrad = self.manifold.proj_tangent(x, grad)
        rgrad_old = rgrad  

        x_local_average = x

        rgrad_average = np.empty(x.shape, dtype = np.float64)
        comm.Allreduce(rgrad, rgrad_average, MPI.SUM)
        rgrad_average = rgrad_average / size
        initial_kkt_violation = np.linalg.norm(rgrad_average, ord = 'fro')

        kkt_violation = initial_kkt_violation / initial_kkt_violation
        consensus_error = 1

        comm.Barrier()

        if self.log is not None:
            start_time = MPI.Wtime()
            self.log.record(start_time = start_time)
            self.log.record(step_size = self.step_size)
            self.log.record(consensus_step = self.consensus_step)
            self.log.record(iter = iter)
            self.log.record(kkt_violation = kkt_violation)
            self.log.record(consensus_error = consensus_error)
            if self.if_compute_obj_value:
                fval_sum = comm.allreduce(fval, MPI.SUM)
                self.log.record(function_value = fval_sum)

        comm.Barrier()

        for iter in range(1, max_iter + 1):

            x_pre = x
            x_local_average_pre = x_local_average
            rgrad_pre = rgrad

            inner = np.int(np.ceil(self.local_obj.sample_size / self.batch_size))
            sample = np.random.permutation(self.local_obj.sample_size)
            end = 0

            #step_size_grad_iter = step_size_grad / math.sqrt(iter)
            step_size_grad_iter = step_size_grad / iter
            
            """ stochastic gradient step, one epoch """
            for _ in range(inner):  # all agents should have same iterations to aviod deadlock

                start = end
                end = start + self.batch_size
                if end > self.local_obj.sample_size:
                    end = self.local_obj.sample_size
                idx = sample[start : end]

                fval, grad = self.local_obj.objective(x, idx, self.if_compute_obj_value)

                # multi-step consensus
                comm_result = self.consensus.communication(x, self.consensus_step)
                x_local_average = comm_result[ : x.size].reshape(self.initial_guess.shape)

                grad_consensus = self.manifold.proj_tangent(x, x_local_average - x)

                # update Riemannian gradient
                rgrad = self.manifold.proj_tangent(x, grad)

                """  update variable  """
                x = self.manifold.polar_retr(x, step_size_cons * grad_consensus - step_size_grad_iter * rgrad)

                if self.log is not None:
                    self.log.record(kkt_violation = kkt_violation)
                    self.log.record(consensus_error = consensus_error)
                    if self.if_compute_obj_value:
                        fval_sum = comm.allreduce(fval, MPI.SUM)
                        self.log.record(function_value = fval_sum)

            fval, grad = self.local_obj.objective(x, np.arange(mi), self.if_compute_obj_value)
            rgrad = self.manifold.proj_tangent(x, grad)

            # compute KKT violation
            rgrad_average = np.empty(x.shape, dtype = np.float64)
            comm.Allreduce(rgrad, rgrad_average, MPI.SUM)
            rgrad_average = rgrad_average / size
            kkt_violation = np.linalg.norm(rgrad_average, ord = 'fro') ** 2
            kkt_violation = kkt_violation / initial_kkt_violation

            # compute consensus error
            x_average = np.empty(x.shape, dtype = np.float64)
            comm.Allreduce(x, x_average, MPI.SUM)
            # u, s, vh = np.linalg.svd(x_average, full_matrices = False)
            # x_average_orth = u @ vh
            x_average_orth = x_average / size
            local_distance_to_average = np.linalg.norm(x_average_orth - x, ord = 'fro') ** 2   
            #consensus_error = np.empty([1, 1], dtype = np.float64)
            consensus_error = comm.allreduce(local_distance_to_average, MPI.SUM)
            consensus_error = consensus_error / size

            # compute feasibility violation
            local_feasibility_violation = np.linalg.norm(x.T @ x - np.eye(x.shape[1]), ord = 'fro') ** 2
            feasibility_violation = comm.allreduce(local_feasibility_violation, MPI.SUM)
            feasibility_violation = feasibility_violation / size

            kkt_violation = np.sqrt(kkt_violation + consensus_error + feasibility_violation)
            consensus_error = np.sqrt(consensus_error)
            feasibility_violation = np.sqrt(feasibility_violation)

            if self.log is not None:
                self.log.record(iter = iter)

            # if rank == 0:
            #     print(feasibility_violation)

            comm.Barrier()

            # check stop criteria
            if size > 1:
                stop_criteria = (kkt_violation < tol_grad) or (consensus_error < tol_cons)
            else:
                stop_criteria = (kkt_violation < tol_grad)
            
            if stop_criteria:
                break

            # # update BB step size
            # if step_size_type == 'BB':
            #     #s = x - x_pre
            #     s = x_local_average - x_local_average_pre
            #     y = rgrad - rgrad_pre
            #     #y = rgrad - rgrad_pre
            #     step_size_grad = abs(np.sum(y * s)) / (np.linalg.norm(y, ord = 'fro') ** 2)
            #     self.step_size['gradient'] = step_size_grad
            #     self.log.record(step_size = self.step_size)

            
            # if rank == 0:
            #     print(step_size_grad)
        
        self.solution = x
        return x
