""" Decentralized Optimization Over the Stiefel Manifold by an Approximate Augmented Lagrangian Function"""


__author__ = 'Lei Wang'
__email__ = 'wlkings@lsec.cc.ac.cn'


import numpy as np
from mpi4py import MPI

from Log_File.Log import Log


comm = MPI.COMM_WORLD
size = MPI.COMM_WORLD.Get_size()
rank = MPI.COMM_WORLD.Get_rank()

class VRSGT:

    def __init__(self, 
                 local_obj,
                 consensus,
                 initial_guess, 
                 penalty_param,
                 batch_size,
                 step_size, 
                 consensus_step,
                 termination_condition = None,
                 log = None, 
                 if_compute_obj_value = False):
    
        self.local_obj = local_obj
        self.consensus = consensus
        self.initial_guess = initial_guess
        self.penalty_param = penalty_param
        self.batch_size = batch_size
        self.step_size = step_size
        self.consensus_step = consensus_step
        self.termination_condition = termination_condition
        self.log = log
        self.if_compute_obj_value = if_compute_obj_value

    def minimize(self):

        iter = 0

        mi = self.local_obj.sample_size
        idx = np.arange(mi)

        assert isinstance(self.batch_size, int) and self.batch_size > 0, 'batch size should be a positive integer'

        max_iter = self.termination_condition.get('iteration')
        if max_iter is None:
            max_iter = 10**4
        tol_grad = self.termination_condition.get('gradient_tolerance')
        if tol_grad is None:
            tol_grad = 1e-4
        tol_cons = self.termination_condition.get('consensus_tolerance')
        if tol_cons is None:
            tol_cons = 1e-10
        stop_criteria = iter < max_iter

        step_size_cons = self.step_size.get('consensus')
        if step_size_cons is None:
            step_size_cons = 1
        step_size_grad = self.step_size.get('gradient')
        if step_size_grad is None:
            step_size_grad = 0.0001
        step_size_type = self.step_size.get('type')
        if step_size_type is None:
            step_size_type = 'BB'

        penalty_param = self.penalty_param / size
        if self.log is not None:
            self.log.record(penalty_param = self.penalty_param)

        x = self.initial_guess

        """  full gradient  """
        xTx = x.T @ x
        xxTx = x @ xTx
        y = x
        fval, grad = self.local_obj.objective(y, np.arange(mi), self.if_compute_obj_value)
        Lambda = x.T @ grad
        Lambda = (Lambda + Lambda.T) / 2
        rgrad = grad @ (1.5 * np.eye(x.shape[1]) - 0.5 * xTx) - x @ Lambda
        D = rgrad + penalty_param * (xxTx - x)
        D_old = D
        D_tracking = D   

        svrg = grad

        rgrad_average = np.empty(x.shape, dtype = np.float64)
        comm.Allreduce(rgrad, rgrad_average, MPI.SUM)
        rgrad_average = rgrad_average / size
        initial_kkt_violation = np.linalg.norm(rgrad_average, ord = 'fro') ** 2

        kkt_violation = initial_kkt_violation / initial_kkt_violation
        consensus_error = 1
        feasibility_violation = 1

        comm.Barrier()

        # # multi-step consensus
        # comm_vector = np.append(x, D_tracking)
        # #comm_result = consensus(comm_vector, peer, self.weight, self.consensus_step)
        # comm_result = self.consensus.communication(comm_vector, self.consensus_step)

        # # update consensus step
        # x_local_average = comm_result[ : x.size].reshape(self.initial_guess.shape)
        # D_tracking = comm_result[x.size : ].reshape(self.initial_guess.shape)


        if self.log is not None:
            start_time = MPI.Wtime()
            self.log.record(start_time = start_time)
            self.log.record(step_size = self.step_size)
            self.log.record(consensus_step = self.consensus_step)
            self.log.record(iter = iter)
            self.log.record(kkt_violation = kkt_violation)
            self.log.record(kkt_violation = kkt_violation)
            self.log.record(consensus_error = consensus_error)
            self.log.record(consensus_error = consensus_error)
            if self.if_compute_obj_value:
                fval_sum = comm.allreduce(fval, MPI.SUM)
                self.log.record(function_value = fval_sum)

        comm.Barrier()

        for iter in range(1, max_iter + 1):

            inner = np.int(np.ceil(self.local_obj.sample_size / self.batch_size))
            sample = np.random.permutation(self.local_obj.sample_size)
            end = 0

            """ stochastic gradient step, one epoch """
            for _ in range(inner):  # all agents should have same iterations to aviod deadlock

                x_pre = x
                y_pre = y
                xTx_pre = xTx
                xxTx_pre = xxTx
                svrg_pre = svrg
                D_pre = D
                D_tracking_pre = D_tracking

                yy = x - step_size_grad * D_tracking

                yy_norm = np.linalg.norm(yy, ord = 'fro')
                radius = 1.01 * np.sqrt(x.shape[1])
                if yy_norm > radius:
                    yy = radius * yy / yy_norm

                # normalize each column
                # yy_norm_column = np.sqrt((yy * yy).sum(axis = 0))
                # yy = yy / yy_norm_column

                # alpha = 1 / iter
                alpha = 1
                z = x + alpha * (yy - x)

                # multi-step consensus
                comm_vector = np.append(z, D_tracking)
                #comm_result = consensus(comm_vector, peer, self.weight, self.consensus_step)
                comm_result = self.consensus.communication(comm_vector, self.consensus_step)

                x = comm_result[ : x.size].reshape(self.initial_guess.shape)

                start = end
                end = start + self.batch_size
                if end > self.local_obj.sample_size:
                    end = self.local_obj.sample_size
                idx = sample[start : end]
               
                # stochastic variance reduced gradient
                xTx = x.T @ x
                xxTx = x @ xTx
                y = x
                fval, grad = self.local_obj.objective(y, idx, self.if_compute_obj_value)
                fval_pre, grad_pre = self.local_obj.objective(xxTx_pre, idx, self.if_compute_obj_value)
                
                svrg = 1 / self.batch_size * (grad - grad_pre) + svrg_pre

                Lambda = x.T @ svrg
                Lambda = (Lambda + Lambda.T) / 2
                rgrad = svrg @ (1.5 * np.eye(x.shape[1]) - 0.5 * xTx) - x @ Lambda
                D = rgrad + penalty_param * (xxTx - x)

                # update Riemannian gradient tracking estimate
                D_tracking = comm_result[x.size : ].reshape(self.initial_guess.shape)
                D_tracking += D - D_old

                # update old Riemannian gradient
                D_old = D

                # if rank == 0:
                #     print(feasibility_violation)

                if self.log is not None:
                    self.log.record(kkt_violation = kkt_violation)
                    self.log.record(kkt_violation = kkt_violation)
                    self.log.record(feasibility_violation = feasibility_violation)
                    self.log.record(feasibility_violation = feasibility_violation)
                    self.log.record(consensus_error = consensus_error)
                    self.log.record(consensus_error = consensus_error)
                    if self.if_compute_obj_value:
                        fval_sum = comm.allreduce(fval, MPI.SUM)
                        self.log.record(function_value = fval_sum)


            x_pre = x
            #x_local_average_pre = x_local_average
            D_pre = D
            D_tracking_pre = D_tracking

            yy = x - step_size_grad * D_tracking

            yy_norm = np.linalg.norm(yy, ord = 'fro')
            radius = 1.01 * np.sqrt(x.shape[1])
            if yy_norm > radius:
                yy = radius * yy / yy_norm

            # normalize each column
            # yy_norm_column = np.sqrt((yy * yy).sum(axis = 0))
            # yy = yy / yy_norm_column

            # alpha = 1 / iter
            alpha = 1
            z = x + alpha * (yy - x)

            # multi-step consensus
            comm_vector = np.append(z, D_tracking)
            #comm_result = consensus(comm_vector, peer, self.weight, self.consensus_step)
            comm_result = self.consensus.communication(comm_vector, self.consensus_step)

            x = comm_result[ : x.size].reshape(self.initial_guess.shape)

            """  full gradient  """
            xTx = x.T @ x
            xxTx = x @ xTx
            y = x
            fval, grad = self.local_obj.objective(y, np.arange(mi), self.if_compute_obj_value)
            Lambda = x.T @ grad
            Lambda = (Lambda + Lambda.T) / 2
            rgrad = grad @ (1.5 * np.eye(x.shape[1]) - 0.5 * xTx) - x @ Lambda
            D = rgrad + penalty_param * (xxTx - x)

            svrg = grad

            comm.Barrier()

            # update Riemannian gradient tracking estimate
            D_tracking = comm_result[x.size : ].reshape(self.initial_guess.shape)
            D_tracking += D - D_old

            # update old Riemannian gradient
            D_old = D

            # compute KKT violation
            rgrad_average = np.empty(x.shape, dtype = np.float64)
            comm.Allreduce(rgrad, rgrad_average, MPI.SUM)
            rgrad_average = rgrad_average / size
            kkt_violation = np.linalg.norm(rgrad_average, ord = 'fro') ** 2
            kkt_violation = kkt_violation / initial_kkt_violation

            # compute consensus error
            x_average = np.empty(x.shape, dtype = np.float64)
            comm.Allreduce(x, x_average, MPI.SUM)
            # u, s, vh = np.linalg.svd(x_average, full_matrices = False)
            # x_average_orth = u @ vh
            x_average_orth = x_average / size
            local_distance_to_average = np.linalg.norm(x_average_orth - x, ord = 'fro') ** 2   
            #consensus_error = np.empty([1, 1], dtype = np.float64)
            consensus_error = comm.allreduce(local_distance_to_average, MPI.SUM)
            consensus_error = consensus_error / size

            # compute feasibility violation
            local_feasibility_violation = np.linalg.norm(x.T @ x - np.eye(x.shape[1]), ord = 'fro') ** 2
            feasibility_violation = comm.allreduce(local_feasibility_violation, MPI.SUM)
            feasibility_violation = feasibility_violation / size

            # if rank == 0:
            #     print(kkt_violation)

            kkt_violation = np.sqrt(kkt_violation + consensus_error + feasibility_violation)
            consensus_error = np.sqrt(consensus_error)
            feasibility_violation = np.sqrt(feasibility_violation)
            
            if self.log is not None:
                self.log.record(iter = iter)
                self.log.record(kkt_violation = kkt_violation)
                self.log.record(kkt_violation = kkt_violation)
                self.log.record(feasibility_violation = feasibility_violation)
                self.log.record(feasibility_violation = feasibility_violation)
                self.log.record(consensus_error = consensus_error)
                self.log.record(consensus_error = consensus_error)
                if self.if_compute_obj_value:
                    fval_sum = comm.allreduce(fval, MPI.SUM)
                    self.log.record(function_value = fval_sum)

            # check stop criteria
            if size > 1:
                stop_criteria = (kkt_violation < tol_grad) or (consensus_error < tol_cons)
            else:
                stop_criteria = (kkt_violation < tol_grad)
                
            if stop_criteria:
                break

            # update BB step size
            # if step_size_type == 'BB':
            #     s = x - x_pre
            #     #s = x_local_average - x_local_average_pre
            #     r = D_tracking - D_tracking_pre
            #     #r = D - D_pre
            #     step_size_grad = abs(np.sum(r * s)) / (np.linalg.norm(r, ord = 'fro') ** 2)
            #     self.step_size['gradient'] = step_size_grad
            #     self.log.record(step_size = self.step_size)

            
            # if rank == 0:
            #     print(step_size_grad)
        
        u, s, vh = np.linalg.svd(x, full_matrices = False)
        x_orth = u @ vh
        self.solution = x_orth
        return x_orth
