import numpy as np

def generate_synthetic_data(rows, cols, eigengap):
    """ generate synthetic data """
    #np.random.seed(seed = 100)
    A = np.random.randn(rows, cols)
    u, singularvalue, vh = np.linalg.svd(A, full_matrices = False)

    top_s = singularvalue[0]
    #top_s = 1
    x_opt = vh
    singularvalue = np.array([top_s * eigengap ** (i / 2) for i in range(cols)])

    # top = np.array([top_s * 0.999 ** (i / 2) for i in range(10)])
    # remaining = np.array([top[9] * eigengap ** (i / 2) for i in range(cols - 10)])
    # singularvalue = np.hstack((top, remaining))

    data = np.matmul(u, vh * singularvalue[:, None])
    return data

def partition_data(data, rows, cols, if_sparse_mat = False):
    """ partition data into n folds """

    from mpi4py import MPI
    comm = MPI.COMM_WORLD
    size = MPI.COMM_WORLD.Get_size()
    rank = MPI.COMM_WORLD.Get_rank()

    # rows = data.shape[0]
    # cols = data.shape[1]
    count = rows // size
    remainder = rows % size

    if not if_sparse_mat:
        if rank == 0:
            if rank < remainder:
                # The first 'remainder' ranks get 'count + 1' tasks each
                start = 0
                stop = start + count
            else:
                # The remaining 'size - remainder' ranks get 'count' task each
                start = 0
                stop = start + (count - 1)
            local_data = data[start : stop + 1, :]
            #print(graph.W)
            #weighted = graph.W[0, :]

            # divide global_data into local data
            for i in range(1, size):
                if i < remainder:
                    # The first 'remainder' ranks get 'count + 1' tasks each
                    start = i * (count + 1)
                    stop = start + count
                else:
                    # The remaining 'size - remainder' ranks get 'count' task each
                    start = stop + 1
                    stop = start + (count - 1)
                #sendbuf = np.append(data[start:stop + 1, :], graph.W[i, :])
                sendbuf = data[start : stop + 1, :]
                comm.Send(sendbuf, dest = i)

        elif rank < remainder:
            recvbuf = np.empty((count + 1) * cols)
            comm.Recv(recvbuf, source = 0)
            local_data = recvbuf[ : (count + 1) * cols].reshape(count + 1, cols)
        else:
            recvbuf = np.empty(count * cols)
            comm.Recv(recvbuf, source = 0)
            local_data = recvbuf[ : count * cols].reshape(count, cols)

        return local_data

    else:
        """  partition for csr sparse matrix:
        for the i-th row, the column indices of non-zeros are indices[indptr[i]:indptr[i+1]],
                        data is data[indptr[i]:indptr[i+1]] """
        from scipy.sparse import csr_matrix

        def find_idx(arr, begin, end):
            if arr[begin] == arr[end]:
                return 0, 0
            return arr[begin], arr[end]

        def reform_row_num(arr, lenth):
            row_n = np.zeros(lenth)
            idx = 0
            x_old = arr[0]
            for i in range(1, lenth):
                rep = arr[i] - x_old
                for j in range(rep):
                    row_n[idx+j] = i-1
                x_old = arr[i]
                idx += rep
                if idx >= lenth: break
            return row_n

        if rank == 0:
            csr_l = [data.data, data.indices, data.indptr]
            if rank < remainder:
                # The first 'remainder' ranks get 'count + 1' tasks each ( row no. start: stop+1)
                start = 0
                stop = start + count
            else:
                # The remaining 'size - remainder' ranks get 'count' task each
                start = 0
                stop = start + (count - 1)

            ind_start, ind_end = find_idx(csr_l[2], start, stop + 1)
            _data = csr_l[0][ind_start : ind_end]
            col_num = csr_l[1][ind_start : ind_end]
            row_num = reform_row_num(csr_l[2][start : stop + 2], col_num.size)
            local_data = csr_matrix((_data, (row_num, col_num)), shape = (stop - start + 1, cols))

            # divide global_data into local data
            for i in range(1, size):
                if i < remainder:
                    # The first 'remainder' ranks get 'count + 1' tasks each
                    start = stop + 1
                    stop = start + count
                else:
                    # The remaining 'size - remainder' ranks get 'count' task each
                    start = stop + 1
                    stop = start + (count - 1)

                ind_start, ind_end = find_idx(csr_l[2], start, stop + 1)
                _data = csr_l[0][ind_start : ind_end]
                col_num = csr_l[1][ind_start : ind_end]

                row_num = reform_row_num(csr_l[2][start : stop + 2], col_num.size)
                # local_data = csr_matrix((_data, (row_num, col_num)), shape=(stop - start + 1, dimension))

                comm.send(row_num.size, dest = i, tag = 1)
                sendbuf = np.concatenate((row_num, col_num, _data), axis = None)
                comm.Send(sendbuf, dest = i, tag = 2)

        elif rank < remainder:
            recvsize = comm.recv(source = 0, tag = 1)
            recvbuf = np.empty(recvsize * 3)
            comm.Recv(recvbuf, source = 0, tag = 2)
            row_num = recvbuf[ : recvsize]
            col_num = recvbuf[recvsize : 2 * recvsize]
            _data = recvbuf[2 * recvsize : 3 * recvsize]
            local_data = csr_matrix((_data, (row_num, col_num)), shape = (count + 1, cols))
        else:
            recvsize = comm.recv(source = 0, tag = 1)
            recvbuf = np.empty(recvsize * 3)
            comm.Recv(recvbuf, source = 0, tag = 2)
            row_num = recvbuf[ : recvsize]
            col_num = recvbuf[recvsize : 2 * recvsize]
            _data = recvbuf[2 * recvsize : 3 * recvsize]
            local_data = csr_matrix((_data, (row_num, col_num)), shape = (count, cols))

        return local_data