from scipy import sparse
import numpy as np


from mpi4py import MPI
comm = MPI.COMM_WORLD
size = MPI.COMM_WORLD.Get_size()
rank = MPI.COMM_WORLD.Get_rank()

class PCA:
    """"
     Quadratic function:  f_i(x) =  -  0.5* sum_{j=1}^sample_size (x'*a_i'*a_i*x)
       #:param sample_size: number of instances at local node
               dimension: dimension of variable
               m: maximum number of instances across all nodes, only needed for decentralized alg
               data: matrix with size (sample_size, dimension)
               idx: row index of A and y
               eigengap: eigengap of matrix A: lambda_i/ lambda_{i+1}= eigengap
               sparse_mat: whether A is a csr sparse matrix
               x_opt: the optimal solution
       #:return a sub-gradient vector:  partial f(x)
       #:return obj val: f(x)
      """
    def __init__(self, data = np.empty([1,1]), if_sparse_mat = False):
        self.data = data
        self.sample_size = data.shape[0]
        self.if_sparse_mat = if_sparse_mat
        self.linear_term = None
        self.regularizer_parameter = None

    def objective(self, x, idx, if_compute_obj_value = False):
        """"
            :param return_objval: whether return objective value
            :param sparse_mat: whether A is a csr sparse matrix
            :return gradient: A'*A*x
            :return obj val: 0.5*Tr(x'*A'*A*x)
        """
        full_idx = np.arange(self.sample_size)
        if not self.if_sparse_mat:
            if idx.size == full_idx.size and (idx == full_idx).all():
                """full grad """
                Ax = np.matmul(self.data, x)
                grad = -np.matmul(self.data.T, Ax)
            elif idx.size == 0:
                """ empty idx """
                return None, np.zeros(x.shape)
            else:
                """ batch grad """
                Ax = np.matmul(self.data[idx, :], x)  # m*1
                grad = -np.matmul(self.data[idx, :].T, Ax)
            if if_compute_obj_value is True:
                fval = -0.5 * np.sum(Ax * Ax)
            else:
                fval = None
            return fval, grad
        else:
            " sparse matrix "
            if idx.size == full_idx.size and (idx == full_idx).all():
                """full grad """
                Ax = sparse.csr_matrix.dot(self.data, x)
                grad = -sparse.csr_matrix.dot(self.data.T, Ax)
            elif idx.size == 0:
                """ empty idx """
                return None, np.zeros(x.shape)
            else:
                """ batch grad """
                if idx.size == 1:
                    idx = int(idx)
                    d1 = self.data.data[self.data.indptr[idx] : self.data.indptr[idx + 1]]
                    i1 = self.data.indices[self.data.indptr[idx] : self.data.indptr[idx + 1]]
                    A_dense = np.zeros(x.size)
                    A_dense[i1] = d1
                    Ax = np.dot(A_dense, x)
                    grad = -np.dot(A_dense.T, Ax)
                else:
                    A1 = self.data[idx, :]
                    Ax = sparse.csr_matrix.dot(A1.T, x)
                    grad = -sparse.csr_matrix.dot(A1.T, Ax)
            if if_compute_obj_value is True:
                fval = -0.5 * np.sum(Ax * Ax)
            else:
                fval = None
            return fval, grad