import pickle
import numpy as np
from mpi4py import MPI

# import os, sys
# sys.path.append("..") 

import os, sys 
parentdir = os.path.dirname(os.path.dirname(os.path.abspath(__file__))) 
sys.path.insert(0, parentdir)

from Algorithm.DRSGD import DRSGD
from Algorithm.VRSGT import VRSGT

from Communication.Graph import Graph
from Communication.Consensus import Consensus

from Objective.PCA import PCA
from Objective.Data import generate_synthetic_data
from Objective.Data import partition_data

from Manifold_Toolbox.Euclidean_Space import Euclidean_Space
from Manifold_Toolbox.Stiefel_Manifold import Stiefel_Manifold

from Log_File.Log import Log

comm = MPI.COMM_WORLD
size = MPI.COMM_WORLD.Get_size()
rank = MPI.COMM_WORLD.Get_rank()


if __name__ == '__main__':

    graph_type = ['Ring', 'ER', 'Star', 'Complete']
    weighted_type = ['Laplacian_Based', 'Lazy_Metropolis', 'Metropolis_Hastings']
    ER_probability = 0.5
    graph_set = (graph_type[1], weighted_type[2], ER_probability)

    sample_size = 1 * size * 10 ** 3
    var_rows = 1 * 10 ** 2
    var_cols = 5
    eigengap = 0.8
    data_size = (sample_size, var_rows, var_cols, eigengap)

    batch_size = 10

    if rank == 0:
        """"create global model """
        global_data = generate_synthetic_data(sample_size, var_rows, eigengap)
        # import scipy.linalg as la
        # _, _, vh = la.svd(global_data, full_matrices = False)
        # x_opt = np.ascontiguousarray(vh[ : var_cols].T)
        _, S, _ = np.linalg.svd(global_data)

        """"create communication network """
        graph = Graph(graph_type = graph_set[0], weighted_type = graph_set[1], N = size, p = graph_set[2], if_plot_graph = False)
        weighted_matrix = graph.generate_weighted_matrix()

        print('===========================================  Test Instance  ===========================================')
        print("data matrix shape:", (sample_size, var_rows), " || ", "variable shape:", (var_rows, var_cols), " || ", "eigengap:", eigengap, " || ", "batch size:", batch_size)
        print("graph type:", graph_set[0], " || ", "weighted matrix type:", graph_set[1], " || ", "probability:", ER_probability, " || ", "size:", size)
        print(S[ : var_cols])
        # print("The peers of graph:", graph.peer)
        # print(graph.W)
    else:
        global_data = None
        weighted_matrix = None
        x_opt = np.empty([var_rows, var_cols], dtype = np.float64)

    """ partition global data into local data """
    local_data = partition_data(data = global_data, rows = sample_size, cols = var_rows, if_sparse_mat = False)

    weight = np.empty(size, dtype = np.float64)
    comm.Scatter(weighted_matrix, weight, root = 0)
    peer = []
    for i in range(size):
        if weight[i] > 0 and i != rank:
            peer.append(i)
    consensus = Consensus(peer = peer, weight = weight)

    """ create local model using partitioned data """
    local_obj = PCA(data = local_data, if_sparse_mat = False)

    #np.random.seed(seed = 100)
    initial_guess = np.random.randn(var_rows, var_cols)
    initial_guess = Stiefel_Manifold.proj_manifold(initial_guess)
    comm.Bcast(initial_guess, root = 0)

    del global_data




    #==============================================     DRSGD      ==============================================#


    Algorithm = DRSGD
    DRSGD_log = Log()
    manifold = Stiefel_Manifold

    consensus_step = 1
    step_size_cons = 1
    step_size_grad = 0.2
    step_size_type = 'BB'
    step_size_type = 'constant'
    step_size = {'consensus': step_size_cons, 'gradient': step_size_grad, 'type': step_size_type}

    if_compute_obj_value = False

    max_iter = 2 * 10 ** 2
    tol_grad = 1e-16
    tol_cons = 1e-16
    termination_condition = {'iteration': max_iter, 'gradient_tolerance': tol_grad, 'consensus_tolerance': tol_cons}

    if rank == 0:
        print('===========================================      DRSGD      ===========================================')
        DRSGD_log.record(alg_name = Algorithm.__name__, data_shape = data_size, graph = graph_set, size = size)
        print('consensus_step =', consensus_step, ' || ', 'step_size_cons =', step_size_cons, ' || ', 
                'step_size_type:', step_size_type, ' || ', 'step_size_grad =', step_size_grad)


    alg = Algorithm(local_obj = local_obj,
                    manifold = manifold,
                    consensus = consensus,
                    initial_guess = initial_guess,
                    consensus_step = consensus_step,
                    batch_size = batch_size,
                    step_size = step_size,
                    termination_condition = termination_condition,
                    log = DRSGD_log,
                    if_compute_obj_value = if_compute_obj_value)


    files_name = 'Results/Pkls'
    if rank == 0:
        if not os.path.isdir(files_name):
            os.makedirs(files_name)

    save_file_name = os.path.join(files_name, 'DRSGD.pkl')
    filename = save_file_name

    """ run the algorithm """
    start_time = MPI.Wtime()
    x = alg.minimize()
    end_time = MPI.Wtime()

    wtime = end_time - start_time
    wall_clock = 0.0
    wall_clock = comm.allreduce(wtime, op = MPI.MAX)
    DRSGD_log.record(wall_clock = wall_clock)
    
    DRSGD_info = 'DRSGD           %f         %d              %e           %e' %(DRSGD_log.wall_clock, DRSGD_log.end_iter, DRSGD_log.kkt_violation, DRSGD_log.consensus_error)

    # save results
    if rank == 0 and alg.log:
        log_file = alg.log
        with open(filename, 'wb') as output:  
            pickle.dump(log_file, output, pickle.HIGHEST_PROTOCOL)


    #==============================================     VRSGT      ==============================================#


    Algorithm = VRSGT
    VRSGT_log = Log()

    penalty_param = 1

    consensus_step = 1
    step_size_cons = 1
    step_size_grad = 0.0001
    step_size_type = 'BB'
    step_size_type = 'constant' #0.0005
    step_size = {'consensus': step_size_cons, 'gradient': step_size_grad, 'type': step_size_type}

    if_compute_obj_value = False

    max_iter = 1 * 10 ** 2
    tol_grad = 1e-16
    tol_cons = 1e-16
    termination_condition = {'iteration': max_iter, 'gradient_tolerance': tol_grad, 'consensus_tolerance': tol_cons}

    if rank == 0:
        print('===========================================      VRSGT      ===========================================')
        VRSGT_log.record(alg_name = Algorithm.__name__, data_shape = data_size, graph = graph_set, size = size)
        print('consensus_step =', consensus_step, ' || ', 'step_size_cons =', step_size_cons, ' || ', 
                'step_size_type:', step_size_type, ' || ', 'step_size_grad =', step_size_grad)


    alg = Algorithm(local_obj = local_obj,
                    consensus = consensus,
                    initial_guess = initial_guess,
                    penalty_param = penalty_param,
                    consensus_step = consensus_step,
                    batch_size = batch_size,
                    step_size = step_size,
                    termination_condition = termination_condition,
                    log = VRSGT_log,
                    if_compute_obj_value = if_compute_obj_value)


    files_name = 'Results/Pkls'
    if rank == 0:
        if not os.path.isdir(files_name):
            os.makedirs(files_name)

    save_file_name = os.path.join(files_name, 'VRSGT.pkl')
    filename = save_file_name

    """ run the algorithm """
    start_time = MPI.Wtime()
    x = alg.minimize()
    end_time = MPI.Wtime()

    wtime = end_time - start_time
    wall_clock = 0.0
    wall_clock = comm.allreduce(wtime, op = MPI.MAX)
    VRSGT_log.record(wall_clock = wall_clock)
    
    VRSGT_info = 'VRSGT           %f         %d              %e           %e' %(VRSGT_log.wall_clock, VRSGT_log.end_iter, VRSGT_log.kkt_violation, VRSGT_log.consensus_error)

    # save results
    if rank == 0 and alg.log:
        log_file = alg.log
        with open(filename, 'wb') as output:  
            pickle.dump(log_file, output, pickle.HIGHEST_PROTOCOL)
    

    if rank == 0:
        print('===========================================       PCA       ===========================================')
        print('Algorithm         Time         Iteration         KKT Violation         Consensus Error   ')
        
        print(DRSGD_info)
        print(VRSGT_info)


    if rank == 0:
        print('\n')